// mtrand.cpp

#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>

#include "mtrand.h"

//   Period parameters
#define N N_MTRAND_FACTOR
#define M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

// Tempering parameters
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

// This function makes number from time

int Gauss_rand::got_deviate;
double Gauss_rand::extra_deviate;


int seed_by_time()
{
	static int co = 0;
	static int prev_value = 0;
	int value;

#if defined(WIN32_PLATFORM)
	LARGE_INTEGER tt;
	QueryPerformanceCounter(&tt);
	value = (int)tt.LowPart;

#else // UNIX
	struct timezone tz;
	struct timeval tv;
	gettimeofday(&tv, &tz);
	value = (int)tv.tv_usec;
#endif

	co++;
	if(value == prev_value) value += co;
	prev_value = value;
	return value;
}

/* initializing the array with a NONZERO seed */
void MTRand::srand(unsigned long seed)
{
	/* setting initial seeds to mt[N] using         */
	/* the generator Line 25 of Table 1 in          */
	/* [KNUTH 1981, The Art of Computer Programming */
	/*    Vol. 2 (2nd Ed.), pp102]                  */
	mt[0]= seed & 0xffffffff;
	for (mti = 1; mti < N; mti++)
		mt[mti] = (69069 * mt[mti - 1]) & 0xffffffff;
}

MTRand::MTRand(unsigned long seed)
{
	this->srand(seed);
}

void MTRand::srandByTime()
{
	this->srand((unsigned long)seed_by_time());
}

MTRand::MTRand()
{
	this->srand((unsigned long)seed_by_time());
}

double MTRand::genrand() const
{
    unsigned long y;
    static unsigned long mag01[2] = {0x0, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;
/*
        if (mti == N+1)   // if sgenrand() has not been called,
            sgenrand(4357);  // a default initial seed is used
*/
        for (kk = 0; kk < N - M; kk++) {
            y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK);
            mt[kk] = mt[kk + M] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        for (;kk< N - 1; kk++) {
            y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
            mt[kk] = mt[kk + (M - N)] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        y = (mt[N - 1] & UPPER_MASK) | (mt[0] & LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

        mti = 0;
    }

    y = mt[mti++];
    y ^= TEMPERING_SHIFT_U(y);
    y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
    y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
    y ^= TEMPERING_SHIFT_L(y);

    return (double)y * 2.3283064365386963e-10; /* reals: [0,1)-interval */
    /* return y; */ /* for integer generation */
}

int MTRand::genrand_int(int size) const
{
	return (int)(genrand() * (double)size);
}

Gauss_rand::Gauss_rand()
{
	got_deviate = GRAND_FALSE;
}

double Gauss_rand::gen_gauss_rand(double mean, double std_dev)
{
	double v1, v2,R, fac, this_deviate;

	if( got_deviate == GRAND_TRUE ){
		this_deviate = extra_deviate;
		got_deviate = GRAND_FALSE;
	} else {
		R = 0.0;
		do{
			v1 = ( 2.0 * genrand())- 1.0;
			v2 = ( 2.0 * genrand())- 1.0;
			R = ( v1 * v1 ) + ( v2 * v2 );
		} while( (R <= 0.0) || (R >= 1.0) );

		fac = sqrt( ( -2.0*log(R) ) / R );
		v1 *= fac;
		this_deviate = v1;
		v2 *= fac;
		extra_deviate = v2;
		got_deviate = GRAND_TRUE;
	}

       return ( mean + ( this_deviate * std_dev ) );
}

