//-------------------------------------------------------------------
// FILENAME    : ncga_mpi.cpp
// DESCRIPTION : Operating GA  of Parallel NCGA
//-------------------------------------------------------------------
#include <iostream>
#include <mpi.h>
#include <math.h>
#include <stdlib.h>

#include "ncga.h"
#include "ncga_mpi.h"
#include "gene.h"
#include "oneisland.h"


int Ncga_mpi::rank=1;
int Ncga_mpi::procs=1;

Ncga_mpi::Ncga_mpi(int& argc, char** & argv):Ncga()
{
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &procs);
  unsigned long seed = (unsigned long)time(NULL);
  MPI_Bcast(&seed, 1, MPI_INT, 0, MPI_COMM_WORLD);
  mtrand.srand(seed);
}


void Ncga_mpi::pareto_creat()
{
  if(rank!=ROOT)
    return ;
  else
    Ncga::pareto_creat();
}

Ncga_mpi::~Ncga_mpi()
{
  MPI_Finalize();
}

void Ncga_mpi::initGA()
{
  if(rank==ROOT)
    Ncga::initGA();
  else{
    One::init_base();
    send_count=cross.get_need_num();
    int i;
    tmpChild = new Gene*[send_count];
    for(i=0;i<send_count;i++)
      tmpChild[i] = new Gene(variable.chromosome_length,variable.obj_num);
  }
}

void Ncga_mpi::cycle()
{
  if(rank==ROOT){
    all_open_w_file();
    fprintf(fp1,"Ncga_mpi (Fine-Graind model) procs %d \n",procs);
    fprintf(fp1,"child_pop_num %d \n",child_num);
  }
  int loop_count=0;
  do{
    if(rank==ROOT){
      printf("####loop %d(rank %d) start#####\n",loop_count,rank);
      fprintf(fp1,"loop %d\n",loop_count);
      fprintf(fp3,"loop %d\n",loop_count);
      initialize();
//      pareto_creat();
    }
    else{
      set_s_time();
      generation=0;
      life_flag=1;
      func.re_set_func_count();
      func_count=0;
      func_time=0.0;
    }
    f_output();
	terminal_check();
    for(;life_flag;){
      if(rank==ROOT){
	printf("####rank %d:(mcc loop %d)generation %d procs %d(pop_num %d)#####\n",rank,loop_count,generation,procs,pop.get_pop_num());
	f_output();
      }
      generation++;
      next_1_gene();
    }
     f_output();
    loop_count++;
  }while(loop_count<variable.loop_num);
  set_all_e_time();
  if(rank==ROOT)
    finish_write();
  all_open_file_close();
}


void Ncga_mpi::f_output()
{
  if(rank != ROOT)
    return ;
  One::f_output(perfect_pop);
}

void Ncga_mpi::next_1_gene()
{
  int pop_num;
  if(rank==ROOT){
    pop_num=pop.get_pop_num();
    qsort(pop.ap_gene,0,pop_num-1,obj_base);
    obj_base++;
    obj_base = obj_base % variable.obj_num;
    neighbor_mix(pop.ap_gene,0,pop_num);
  }

  real_mcc_model();
  if(rank==ROOT){
    pareto_creat();
    now_pop_re_create(pop,perfect_pop);
  }
  terminal_check();
}

void Ncga_mpi::terminal_check()
{
  variable_gather();
  if(generation>=variable.end_gene||func.get_func_count() >= variable.limit_func_count){
    life_flag=0;
    set_e_time();
    if(rank==ROOT)
      printf("terminal OK generation%d all_func_count %d \n",generation,func.get_func_count());
  }
  terminal_sign();
}

void Ncga_mpi::terminal_sign()
{
  MPI_Bcast(&life_flag,1,MPI_BYTE,ROOT,MPI_COMM_WORLD);
}

void Ncga_mpi::variable_gather()
{
  int dum,r_dum;
  if(rank==ROOT){
    if(generation==0)
      dum=func.get_func_count();
    else
      dum=0;
  }
  else{
    dum = func.get_func_count();
  }
  MPI_Reduce(&dum,&r_dum,1,MPI_INT,MPI_SUM,ROOT,MPI_COMM_WORLD);
  if(rank==ROOT){
    func.set_func_count(r_dum);
  }
}

void Ncga_mpi::real_mcc_model()
{
  if(procs==1)
    sp_version();
  else
    {
      double time;
      time = get_time();
      mp_version();
      func_time += get_time() - time;
    }
}

void Ncga_mpi::mp_version()
{
  int i,j,msgt,size,sum_size,recv_num;
  int s_rank;
  MPI_Status stat[2];
  MPI_Request request[2];
  u8 *rna_recv,dum_dum;
  u8 *rna_send;
  MPI_Barrier(MPI_COMM_WORLD);

  if(rank==ROOT){
    MPI_Request f_req[procs-1];
    MPI_Status f_st[procs-1];    
    int count,limit,s_size,dum,gene_size;
    
    gene_size=pop.ap_gene[0]->GetRnaLength();
    s_size=get_size_simple();
    rna_send=new u8[s_size];
    
    nslc_transcription(rna_send);    
    count=0;
    dum=0;
    limit= pop.get_pop_num()/(send_count);
    
    msgt=2222;
    sum_size=0;

    for(i=0;i<(procs-1);i++){
      size=0;
      for(j=0;j<send_count;j++)
	size += pop.ap_gene[i*send_count+j]->GetRnaLength();
      MPI_Isend(rna_send+sum_size,size,MPI_BYTE,i+1,msgt,MPI_COMM_WORLD,&f_req[i]);
      sum_size += size;
    }
    MPI_Waitall(procs-1,f_req,f_st);
    //    printf("ROOT mcc_first_send(size %d) OK!\n",size);
    count=procs-1;
    do{
      msgt=4444;
      MPI_Probe(MPI_ANY_SOURCE,msgt,MPI_COMM_WORLD,&stat[0]);
      s_rank=stat[0].MPI_SOURCE;
      MPI_Get_count(&stat[0],MPI_BYTE,&size);
      rna_recv = new u8[size];

      MPI_Irecv(rna_recv,size,MPI_BYTE,s_rank,msgt,MPI_COMM_WORLD,&request[0]);
      recv_num=size/gene_size;

      if(count<limit){
	msgt=2222;
	size=0;
	for(j=0;j<send_count;j++)
	  size += pop.ap_gene[count*send_count+j]->GetRnaLength();
	MPI_Isend(rna_send+sum_size,size,MPI_BYTE,s_rank,msgt,MPI_COMM_WORLD,&request[1]);
	sum_size += size;
	MPI_Waitall(2,request,stat);
      }
      else{ 
	MPI_Wait(&request[0],&stat[0]);
	//	printf("rank %d count %d < limit %d\n ",rank,count,limit);
      }
      size=0;
      for(i=0;i<recv_num;i++,dum++){
	next_pop.ap_gene[dum]->reverse_transcription(rna_recv+size);
	size += pop.ap_gene[dum]->GetRnaLength();
      }
      count++;
      delete [] rna_recv;
    }while(count<(limit+procs-1));
    delete [] rna_send;
    msgt=9999;
    for(i=0;i<(procs-1);i++)
      MPI_Isend(&dum_dum,1,MPI_BYTE,i+1,msgt,MPI_COMM_WORLD,&f_req[i]);
    MPI_Waitall(procs-1,f_req,f_st);
    next_pop.set_pop_num(dum);
    pop.copy(next_pop);
  }
  else{
    int count;
    count=0;
    for(;;){
      MPI_Probe(0,MPI_ANY_TAG,MPI_COMM_WORLD,&stat[0]);
      if(stat[0].MPI_TAG != 2222)
	break;
      msgt=stat[0].MPI_TAG;
      MPI_Get_count(&stat[0],MPI_BYTE,&size);
      try{
	rna_recv = new u8[size];
      }catch(bad_alloc &ba){
	printf("\n**********ERROR *************\n");
      }
      MPI_Irecv(rna_recv,size,MPI_BYTE,ROOT,msgt,MPI_COMM_WORLD,&request[0]);
      MPI_Wait(&request[0],&stat[0]);
      size=0;
      for(i=0;i<send_count;i++){
	tmpChild[i]->reverse_transcription(rna_recv+size);
	size += tmpChild[i]->get_size();
      }
      ga_operator();  
      size = 0;
      for(i=0;i<send_count;i++){
	size += tmpChild[i]->GetRnaLength();
      }
      rna_send = new u8[size];
      size = 0;
      for(i=0;i<send_count;i++){
	tmpChild[i]->transcription(rna_send+size);
	size += tmpChild[i]->get_size();
      }
      msgt=4444;
      MPI_Isend(rna_send,size,MPI_BYTE,ROOT,msgt,MPI_COMM_WORLD,&request[0]);
      MPI_Wait(&request[0],&stat[0]);
      count++;
      delete [] rna_recv;
      delete [] rna_send;
    }
    msgt=9999;
    MPI_Recv(&dum_dum,1,MPI_BYTE,ROOT,msgt,MPI_COMM_WORLD,&stat[0]);
  }
}

void Ncga_mpi::all_open_w_file()
{
  if(rank!=ROOT)
    return;
  Ncga::all_open_w_file();
  fprintf(fp1,"Let's start Master-Slave of NCGA procs= %d\n",procs);
}

void Ncga_mpi::all_open_file_close()
{
  if(rank==ROOT)
    One::all_open_file_close();
}

const u8* Ncga_mpi::nslc_transcription(u8 *buffer)
{
  int size=0,i,pop_num;
  i16 dum_u16;  
  pop_num = pop.get_pop_num();
  size =0;
  for(i=0;i<pop_num;i++){
    pop.ap_gene[i]->transcription(buffer+size);
    size += pop.ap_gene[i]->GetRnaLength();
  }
  return buffer;
}

int Ncga_mpi::get_size_simple(Pop& pop)
{
  int i,size=0;

  for(i=0;i<pop.get_pop_num();i++)
    size += pop.ap_gene[i]->GetRnaLength();
  return size;
}

int Ncga_mpi::get_size_simple()
{
  return get_size_simple(pop);
}








