//-------------------------------------------------------------------
// FILENAME    : oneisland.cpp
// DESCRIPTION : Operating GA 
//-------------------------------------------------------------------

#include <stdio.h>
#include <iostream>
#include <math.h>
#include<stdlib.h>
#include<sys/time.h>
#include <fstream>
#include "pop.h"
#include "func.h"
#include "func_numerical.h"
#include "oneisland.h"
#include "mutation.h"
#include "select.h"
#include "cross_over.h"
#include "loadconf.h"
#include "base.h"

int F_out::obj_num=2;

One::One()
{
	set_all_s_time();
	generation=0;
	life_flag=1;
}

void One::initGA()
{
  init_base();
  int dum;
  dum = (int)((double)variable.pop_num+5.0);
  pop.set_pop_num(variable.pop_num);
  pop.init_ap_gene(dum);
}

void One::init_base()
{
  o_file();
  set_parameter();
  pop.set_obj_num(variable.obj_num);
}

void One:: o_file()
{
  set_parameter_default();
  Conffile cf("mof_ga.txt");
  o_file_base(cf);
  parameter_output();
}

void One:: o_file_base(Conffile &cf)
{
  char *dum_char;
  if(cf.read("pop_num")){
    variable.pop_num = atoi(cf.read("pop_num"));
    cout<<"pop_num base OK  "<<variable.pop_num<<"\n";
  }
  if(cf.read("func_number"))
    variable.func_number = atoi(cf.read("func_number"));
  if(cf.read("cross_method"))
    variable.cross_method = atoi(cf.read("cross_method"));
  if(cf.read("mut_method"))
    variable.mut_method = atoi(cf.read("mut_method"));
  if(cf.read("select_method"))
    variable.select_method = atoi(cf.read("select_method"));
  if(cf.read("sharing_need_pop"))
    variable.sharing_need_pop = atoi(cf.read("sharing_need_pop"));
  if(cf.read("cross_r"))
    variable.cross_r = atof(cf.read("cross_r"));
  if(cf.read("loop_num"))
    variable.loop_num = atoi(cf.read("loop_num"));
  if(cf.read("end_gene"))
    variable.end_gene = atoi(cf.read("end_gene"));
  if(cf.read("limit_func_count")){
    variable.limit_func_count = atoi(cf.read("limit_func_count"));
    variable.end_gene = INT_MAX;
  }
  if(cf.read("print_level"))
    variable.print_level = atoi(cf.read("print_level"));
  if(cf.read("plan_num"))
    variable.plan_num = atoi(cf.read("plan_num"));
  if(cf.read("obj_num"))
    variable.obj_num = atoi(cf.read("obj_num"));
  func.set_func_number(variable.func_number,variable.obj_num,variable.plan_num);

  if(cf.read("file_in")){
    variable.file_in = atoi(cf.read("file_in"));
  }
  variable.obj_num=func.get_obj_num();
  variable.plan_num=func.get_plan_num();
  
  variable.chromosome_length=func.get_plan_num()*20;
  
  if(cf.read("mut_r"))
    variable.mut_r = atof(cf.read("mut_r"));
  else
    variable.mut_r = 1.0/variable.chromosome_length;
}


void One::set_parameter_default()
{
  variable.pop_num=100;
  variable.func_number=1;
  variable.cross_method=1;
  variable.mut_method=1;
  variable.select_method=1;
  variable.sharing_need_pop=100;
  variable.cross_r=1.0;
  variable.mut_r=0.1;
  variable.loop_num=1;
  variable.end_gene=250;
  variable.limit_func_count=INT_MAX;
  variable.print_level=1;
  variable.plan_num=10;
  variable.obj_num=2;
  variable.file_in=0;
}

void One::set_parameter()
{
	select.set_select_method(variable.select_method);
	select.set_obj_num(variable.obj_num);
	select.set_select_pop_num(variable.pop_num);
	select.set_sharing_need_pop(variable.sharing_need_pop);
	mut.set_mut_method(variable.mut_method);
	mut.set_mut_r(variable.mut_r);
	mut.set_plan_num(variable.plan_num);
	mut.set_length(variable.chromosome_length);
	cross.set_cross_r(variable.cross_r);
	cross.set_cross_method(variable.cross_method);
	cross.set_plan_num(variable.plan_num);
	cross.set_chromo_len(variable.chromosome_length);
	func.init_Func(variable.func_number);
	f_out.init(variable.obj_num);
	Gene::set_parameter(variable.obj_num,variable.plan_num,variable.chromosome_length);
	Gene::set_func_num(variable.func_number);
	pop.set_plan_num(variable.plan_num);
	pop.set_obj_num(variable.obj_num);
}

void One::parameter_output()
{
	cerr <<"first_pop_num  = " << variable.pop_num<<"\n";
	cerr <<"func_number  = " << variable.func_number<<"\n";
	cerr <<"obj_num  = " << variable.obj_num<<"\n";
	cerr <<"plan_num  = "<< variable.plan_num<<"\n";
	cerr <<"chromo_length  = "<< variable.chromosome_length<<"\n";
	cerr <<"cross_method  = " << variable.cross_method<<"\n";
	cerr <<"mut_method  = " << variable.mut_method<<"\n";
	cerr <<"select_method  = " << variable.select_method<<"\n";
	cerr <<"sharingg_pop  = " << variable.sharing_need_pop<<"\n";
	cerr <<"cross_r = " << variable.cross_r<<"\n";
	cerr <<" mut_r  = " << variable.mut_r<<"\n";
	cerr <<"end_gene  = " << variable.end_gene<<"\n";
	cerr <<"limit_func_count = " << variable.limit_func_count<<"\n";
}


void One::cycle()
{
  all_open_w_file();
  int loop_count=0;
  do{
    initialize();
    fprintf(fp1,"loop %d\n",loop_count);
    fprintf(fp3,"loop %d\n",loop_count);
    f_output();
    terminal_check();
    for(;life_flag;){
      printf("####(loop %d)generation %d start(pop_num %d)#####\n",loop_count,generation,pop.get_pop_num());
      generation++;
      next_1_gene();
    }
    f_output();
    loop_count++;
  }while(loop_count<variable.loop_num);
  set_all_e_time();
  finish_write();
  all_open_file_close();
}


void One::all_open_w_file()
{
	char filename1[50],filename2[50],filename3[50],filename4[50],filename5[50],filename6[50],filename7[30],filename8[30];
	
	sprintf(filename1,"output_base_parameter_moga_%d.txt",variable.func_number);

	if(variable.print_level != 1)
	  sprintf(filename2,"output_error_moga_%d.txt",variable.func_number);

	sprintf(filename3,"output_time_moga_%d.txt",variable.func_number);
	sprintf(filename4,"output_object_value_moga_%d.txt",variable.func_number);

	if(variable.print_level >= 3 )
	  sprintf(filename5,"output_chromo_moga_%d.txt",variable.func_number);

	sprintf(filename8,"output_one_moga_%d.txt",variable.func_number);
	if(variable.print_level!=1)
	  sprintf(filename7,"output_plan_moga_%d.txt",variable.func_number);

	fp1= file_open(filename1);
	if(variable.print_level != 1){
	  fp2= file_open(filename2);
	  fprintf(fp2,"func_count\t error\n");
	}
	fp3= file_open(filename3);
	fp4= file_open(filename4);

	if(variable.print_level >= 3 )
	  fp5= file_open(filename5);

	if(variable.print_level!=1)
	  fp6 = file_open(filename7);
	fp7= file_open(filename8);
	f_set_output();
}

void One::f_set_output()
{
	fprintf(fp1,"first_pop_num %d\n",variable.pop_num);
	fprintf(fp1,"func_number %d\n",variable.func_number);
	fprintf(fp1,"obj_num %d\nplan_num %d\n",variable.obj_num,variable.plan_num);
	fprintf(fp1,"cross_method %d\n",variable.cross_method);
	fprintf(fp1,"mut_method %d\n",variable.mut_method);
	fprintf(fp1,"select_method %d\n",variable.select_method);
	fprintf(fp1,"sharingg_pop %d\n",variable.sharing_need_pop);
	fprintf(fp1,"cross_r %f\n mut_r %f\n",variable.cross_r,variable.mut_r);
	fprintf(fp1,"end_gene %d\n",variable.end_gene);
	fprintf(fp1,"limit_func_count %d \n",variable.limit_func_count);
	fprintf(fp1,"file_in  %d \n",variable.file_in);
	if(variable.print_level!=1)
	  fprintf(fp6,"cycle %d\n",variable.loop_num);
}

void One::initialize()
{
  initialize(pop);
}

void One::initialize(Pop &dum_pop)
{
	set_s_time();
	generation=0;
	func_time=0.0;
	life_flag=1;
	dum_pop.set_pop_num(variable.pop_num);
	func.re_set_func_count();
	func_count=0;

	if(variable.file_in){
	  ifstream fin;
	  fin.open(INITIAL_POPULATION_FILE);
	  if(!fin.is_open()){
	    cerr << __FILE__ << ":(" << __LINE__ << ") can't open file "
		 << INITIAL_POPULATION_FILE << endl;
	    exit(1);
	  }
	  file_init_pop(dum_pop,fin);
	  fin.close();
	}
	else
	  init_pop(dum_pop);
}

void One::init_pop(Pop &dum_pop)
{
  func.pop_create(dum_pop);
}

void One::file_init_pop(Pop &dum_pop,ifstream & fin)
{
  int i,j;
  double variable_s;
  u8 flag=1;
  vector <double> variable_value(variable.plan_num);
  char dum_char[100];

  fin.ignore(INT_MAX ,'\n');
  for(i = 0; i < dum_pop.get_pop_num()&&fin.good(); ++i){
    fin >>dum_char;
    for(j = 0; j < variable.plan_num&&fin.good(); ++j){
      fin >> variable_s;
      variable_value[j]=variable_s;
      if(fin.fail())
	break;
    }    
    dum_pop.ap_gene[i]->set_locus_value(variable_value); 
  }
  if(i !=dum_pop.get_pop_num()){
    func.pop_create(dum_pop.ap_gene,i,dum_pop.get_pop_num());
  }
  func.func(dum_pop);
}

void One::function()
{
  double time;
  time = get_time();

	cout <<"start function"<<endl;

  func.func(pop);
  
  	for(int i=0;i<variable.obj_num;i++){
  		if(i!=(variable.obj_num-1))
			 printf("%f\t",pop.ap_gene[0]->obj[i]);
		else
			printf("%f\n",pop.ap_gene[0]->obj[i]);
	}
  	
  
  double dumy = pop.ap_gene[0]->get_locus_value_p(0);
  cout<<"dumy = "<<dumy<<endl;
  pop.ap_gene[0]->set_locus_value(0,dumy);
  
  dumy = pop.ap_gene[0]->get_locus_value(0);
  cout<<"(copy)dumy = "<<dumy<<endl;
  
  func.func(*pop.ap_gene[0]);
  
  cout<<"after "<<endl;
  	for(int i=0;i<variable.obj_num;i++){
  		if(i!=(variable.obj_num-1))
			 printf("%f\t",pop.ap_gene[0]->obj[i]);
		else
			printf("%f\n",pop.ap_gene[0]->obj[i]);
	}
  
  func_time += get_time() - time;
}


void One::f_output(){
  f_output(pop);
}

void One::f_output(Pop& dum_pop)
{
  select.obj_setting(dum_pop);
  f_out.avg_obj(dum_pop,fp1);

  if(variable.print_level!=1)
    non_dominated_space(dum_pop);
  
  if(variable.print_level>=2){
    fprintf(fp4,"generation %d \n",generation);
    f_out.all_obj(dum_pop,fp4);    
  }
  if(!life_flag){
    f_out.func_count = func_count;
    f_out.all_func_count +=func_count;
    f_out.avg_func_count += func.get_func_count()/(double)variable.loop_num;  
    f_out.avg_generation += get_generation()/(double)variable.loop_num;
    f_out.generation = get_generation();
    f_out.e_s_time =get_e_s_time();
    f_out.avg_e_s_time += f_out.e_s_time/(double)variable.loop_num;
    f_out.all_e_s_time += f_out.e_s_time;
    f_out.func_time = func_time;
    f_out.avg_func_time += func_time/(double)variable.loop_num;
    f_out.gather_s(fp3);
    f_out.avg_obj(dum_pop,fp1);

    select.only_pareto_select(dum_pop);
    select.reduce_same_pop(dum_pop);
    select.new_sort(dum_pop,0);

    if(variable.print_level != 1){
      f_out.f_end(dum_pop,fp1);
      f_out.all_plan(dum_pop,fp6);
    }
    
    fprintf(fp4,"real pareto frontier \n");
    f_out.all_obj(dum_pop,fp4);    

    if(variable.print_level >= 3 ){
      fprintf(fp5," gene %d:real fronteir chromosome ",generation);
      f_out.all_chromo(dum_pop,fp5);
    }
    f_out.all_obj(dum_pop,fp7);
  }
}

void One::non_dominated_space(Pop &dum_pop)
{
  int dum;
  double errer;
  dum = dum_pop.get_pop_num();
  select.only_pareto_check(dum_pop.ap_gene,dum);
  errer = func.real_func(dum_pop);
  fprintf(fp2,"%d\t%f\n",func.get_func_count(),errer);
}

void One::next_1_gene()
{
  cross_over();
  mutation();
  function();
  selection();
  f_output();
  terminal_check();
}

void One::cross_over()
{
  cross.cross_over(pop);
}

void One::mutation()
{
  mut.mut(pop);
}

void One::selection()
{
  select.selection(pop);
}

void One::terminal_check()
{
  if(generation>=variable.end_gene||func.get_func_count()>=variable.limit_func_count){
    life_flag=0;
    func_count=func.get_func_count();
    set_e_time();
    printf("*******finishu(time %f)********* \n",e_time-s_time);
  }
}

void One::finish_write()
{
  f_out.all_e_s_time=get_all_e_s_time();
  f_out.gather(fp3);
}

void One::all_open_file_close()
{
	fclose(fp1);
	if(variable.print_level != 1)
	  fclose(fp2);
	fclose(fp3);
	fclose(fp4);
	if(variable.print_level >= 3 )
	  fclose(fp5);
	if(variable.print_level!=1)
	  fclose(fp6);
	fclose(fp7);
}


F_out::F_out()
{
	avg_obj_value = NULL;
	e_s_time=0.0;
	all_e_s_time=0.0;
	all_func_count=0;
	avg_func_count=0.0;
	all_generation=0;
	avg_generation=0.0;
	avg_e_s_time=0.0;
	avg_func_time=0.0;
}

F_out::F_out(const int& x)
{
	int i;
	obj_num=x;
	avg_obj_value = new double[obj_num];
	for(i=0;i<obj_num;i++)
		avg_obj_value[i]=0.0;

	e_s_time=0.0;
        all_e_s_time=0.0;
        all_func_count=0;
        avg_func_count=0.0;
        all_generation=0;
        avg_generation=0.0;
	avg_e_s_time=0.0;
}

F_out::F_out(const F_out& f)
{
	int i;
	avg_obj_value = new double[obj_num];
	for(i=0;i<obj_num;i++)
                avg_obj_value[i]=f.avg_obj_value[i];
	e_s_time=f.e_s_time;
	all_e_s_time=f.all_e_s_time;
	all_func_count=f.all_func_count;
	avg_func_count=f.avg_func_count;
	all_generation=f.all_generation;
	avg_generation=f.avg_generation;
	avg_e_s_time=f.avg_e_s_time;
}


F_out::~F_out()
{
  if(avg_obj_value)
    delete[] avg_obj_value;
}

void F_out::init(const int& x)
{
	int i;
	obj_num=x;
	avg_obj_value = new double[obj_num];
        for(i=0;i<obj_num;i++)
                avg_obj_value[i]=0.0;
}

void F_out::avg_obj(Pop& a,FILE *fp)
{
	int i,j,pop_num;
	pop_num=a.get_pop_num();
	for(j=0;j<obj_num;j++)
		avg_obj_value[j]=0.0;

	for(i=0;i<pop_num;i++){
	  for(j=0;j<obj_num;j++){
	    avg_obj_value[j] += a.ap_gene[i]->obj[j];
	  }
	}
	for(j=0;j<obj_num;j++)
	  avg_obj_value[j]=avg_obj_value[j]/pop_num;

//	fprintf(fp,"[");
//	for(j=0;j<obj_num;j++){
//		if(j!=(obj_num-1))
//			fprintf(fp,"%f,",avg_obj_value[j]);
//		else
//			fprintf(fp,"%f]",avg_obj_value[j]);
//	}
//	fprintf(fp,"\n");
//	printf("[");
//	for(j=0;j<obj_num;j++){
//		if(j!=(obj_num-1))
//		printf("%f,",avg_obj_value[j]);
//		else
//		printf("%f]",avg_obj_value[j]);
//	}
//	printf("\n");


	for(j=0;j<obj_num;j++){
//	  cout<<j<<":";
//	  printf("[%f,%f] \n",a.set_max_obj(j),a.set_min_obj(j));
		if(j==(obj_num-1)){
		  fprintf(fp,"%f\n",a.set_min_obj(j));			
		}
		else{
		  fprintf(fp,"%f, ",a.set_min_obj(j));
		}
	}
	
	printf("f_i [ min ] \n[");
	for(j=0;j<obj_num;j++){
//	  cout<<j<<":";
//	  printf("[%f,%f] \n",a.set_max_obj(j),a.set_min_obj(j));
		if(j==(obj_num-1)){
		  printf("%f]\n",a.set_min_obj(j));			
		}
		else{
		  printf("%f, ",a.set_min_obj(j));
		}
	}

}

void F_out::avg_obj(Pop& a)
{
	int i,j,pop_num;
	pop_num=a.get_pop_num();
	for(j=0;j<obj_num;j++)
		avg_obj_value[j]=0.0;

	for(i=0;i<pop_num;i++){
		for(j=0;j<obj_num;j++){
			avg_obj_value[j] += a.ap_gene[i]->obj[j];
		}
	}
	for(j=0;j<obj_num;j++)
                avg_obj_value[j]=avg_obj_value[j]/pop_num;

//	printf("[");
//	for(j=0;j<obj_num;j++){
//		if(j!=(obj_num-1))
//		printf("%f,",avg_obj_value[j]);
//		else
//		printf("%f]",avg_obj_value[j]);
//	}
//	printf("\n");

	printf("f_i [ min ] \n[");
	for(j=0;j<obj_num;j++){
//	  cout<<j<<":";
//	  printf("[%f,%f] \n",a.set_max_obj(j),a.set_min_obj(j));
		if(j==(obj_num-1)){
		  printf("%f]\n",a.set_min_obj(j));			
		}
		else{
		  printf("%f, ",a.set_min_obj(j));
		}
	}
}

void F_out::f_end(Pop& a,FILE *fp)
{
	int i,j,k,pop_num;
	double max_fx[obj_num],min_fx[obj_num],value;
	pop_num = a.get_pop_num();
	for(k=0;k<obj_num;k++){
	  max_fx[k]=a.ap_gene[0]->obj[k];
	  min_fx[k]=a.ap_gene[0]->obj[k];
	  for(i=1;i<pop_num;i++){
	    value = a.ap_gene[i]->obj[k];
	    if(max_fx[k]<value)
	      max_fx[k]=value;
	    else if(min_fx[k]>value)
	      min_fx[k]=value;
	  }
	}
	fprintf(fp,"[max,min]:\n");
	for(j=0;j<obj_num;j++){
	  fprintf(fp,"obj=%d:[%f,%f]\n",j,max_fx[j],min_fx[j]);
//	  fprintf(fp,"obj=%d:[%f]\n",j,min_fx[j]);
	}
}



void F_out::all_obj(Pop& a,FILE *fp)
{
	int i;
	int pop_num;
	pop_num = a.get_pop_num();
	for(i=0;i<pop_num;i++){
			a.ap_gene[i]->f_show_obj(fp);
	}
	fprintf(fp,"\n");
}


void F_out::all_chromo(Pop& a,FILE *fp)
{
        int i,pop_num;

	pop_num=a.get_pop_num();

        fprintf(fp," geneation %d \n",generation);
        for(i=0;i<pop_num;i++){
	  a.ap_gene[i]->f_show_obj(fp);
	  fprintf(fp,"\n");
	  a.ap_gene[i]->f_chromosome(fp);
	}
	fprintf(fp,"\n");
}

void F_out::all_plan(Pop& a,FILE *fp)
{
        int i,pop_num;
	pop_num=a.get_pop_num();
        fprintf(fp," geneation %d \n",generation);
        for(i=0;i<pop_num;i++){
	  a.ap_gene[i]->f_show_plan_value(fp);
	}
	fprintf(fp,"\n");
}


void F_out::gather(FILE *fp)
{
  fprintf(fp,"avg_e_s_time=%f\n",avg_e_s_time);
  fprintf(fp,"avg_func_time=%f \t other time %f\n",avg_func_time,avg_e_s_time - avg_func_time);
  fprintf(fp,"avg_func_count=%f\n",avg_func_count);
  fprintf(fp,"avg_generation=%f\n",avg_generation);
  fprintf(fp,"all_e_s_time=%f\n",all_e_s_time);
}

void F_out::gather_s(FILE *fp)
{
  fprintf(fp,"e_s_time=%f\n",e_s_time);
  fprintf(fp,"func_time %f \t other_time %f\n",func_time, e_s_time-func_time);
  fprintf(fp,"func_count=%d\n",func_count);
  fprintf(fp,"generation=%d\n",generation);
}

FILE* file_open(char *filename)
{
  FILE *fp;
  if((fp=fopen(filename,"w"))==NULL){
    printf("FILE( %s ) not open\n",filename);
    exit(1);
  }
  return fp;
}

double get_time()
{
  struct timeval tv1;
  struct timezone tz;
  double time;
  
  gettimeofday(&tv1,&tz);
  time=(double)tv1.tv_sec+((double)tv1.tv_usec*(1.0E-6));
  return(time);
}


u8 **alloc_carray2(size_t nrow, size_t ncol)
{
  size_t i;
  u8 **m;
  m =  (u8**)malloc( sizeof(char*) * nrow );
  if( m == NULL)
    printf(" error malloc at alloc_carray2. \n");
  
  m[0]= (u8*)malloc( sizeof(char) * nrow * ncol );
  if(m[0] == NULL)
    printf(" error malloc at alloc_carray2. \n");
  
  for( i=1 ; i<nrow ; i++)
    m[i] = m[i-1] + ncol;
  
  for( i=0 ; i < (nrow*ncol) ; i++)
    *(m[0] + i) = 0;
   
  return m;
}

double **alloc_darray2(size_t nrow, size_t ncol)
{
  size_t i;
  double **m;
  
  m =  (double**)malloc( sizeof(double*) * nrow );
  if( m == NULL)
    printf(" error malloc at alloc_darray2. \n");
  
  m[0]= (double*)malloc( sizeof(double) * nrow * ncol );
  if(m[0] == NULL)
    printf(" error malloc at alloc_darray2. \n");
  
  for( i=1 ; i<nrow ; i++)
    m[i] = m[i-1] + ncol;
  
  for( i=0 ; i < (nrow*ncol) ; i++)
    *(m[0] + i) = 0.0;
  
  return m;
}

void free_carray2(u8 **m)
{
  if( m[0] ) free( m[0] );
  if( m ) free( m );
}

void free_darray2(double **m)
{
  if( m[0]  ) free(m[0]);
  if( m ) free(m);
}
