//-------------------------------------------------------------------
// FILENAME    : oneisland.h
// DESCRIPTION : Operating GA 
//-------------------------------------------------------------------
#ifndef _ONE_H_
#define _ONE_H_

#include <fstream>
#include "pop.h"
#include "gene.h"
#include "mutation.h"
#include "cross_over.h"
#include "select.h"
#include "func.h"
#include "func_numerical.h"
#include "loadconf.h"
#include "base.h"

#define INITIAL_POPULATION_FILE "design_variable.ini"

extern FILE *file_open(char *);
extern double get_time();
extern double **alloc_darray2(size_t nrow, size_t ncol);
extern void free_darray2(double **m);

typedef struct  // Configration Date
{
  int pop_num;
  int obj_num;
  int plan_num;
  int chromosome_length;
  double cross_r;
  int cross_method;
  int length;
  double mut_r;
  int mut_method;
  int func_number;
  int select_method;
  int sharing_need_pop;
  int end_gene;
  int limit_func_count;
  int loop_num;
  int print_level;
  int file_in;
}variable_str;


// For outputting date to file
class F_out{ 
public:
        static int obj_num;
        double *avg_obj_value;
        double e_s_time;
	double func_time,avg_func_time;
        double avg_e_s_time,all_e_s_time;
        int all_func_count,func_count;
        double avg_func_count;
        int all_generation,generation;
        double avg_generation;
        F_out();
        F_out(const int&);
        F_out(const F_out&);
        ~F_out();
        void avg_obj(Pop& ,FILE *);
	void avg_obj(Pop&);
	void all_obj(Pop& ,FILE *);
	void f_end(Pop& ,FILE *);
        void gather(FILE *);
        void gather_s(FILE *);
        inline void set_e_s_time(const double& x)       {e_s_time=x;}
	inline void set_obj_num(const int& x)	{obj_num=x;}
	void init(const int& x);
	void all_chromo(Pop& a,FILE *fp);
	void all_plan(Pop& a,FILE *fp);
};

// The outline of GA operation.
class One{
protected:
  mutable MTRand mtrand;
  double avg_generation;
  int func_count;
  double s_time,e_time,all_s_time,all_e_time;
  double func_time;
  Numerical_func func;
  Cross cross;
  F_out f_out;
  Mutation mut;
  Select select;
  variable_str variable;
  Pop pop;
  int generation;
  FILE *fp1,*fp2,*fp3,*fp4,*fp5,*fp6,*fp7;
  u8 life_flag;

  //The following functions are in connection with initGA()
  void init_base();
  virtual void o_file();
  void o_file_base(Conffile &cf);
  void set_parameter_default();
  virtual void set_parameter();
  virtual void parameter_output();
  //The mentioned above functions are in connection with initGA()

  //the following functions are in connection with cycle()
  virtual void all_open_w_file();
  virtual void f_set_output();
  virtual void initialize();
  void initialize(Pop &);
  void file_init_pop(Pop &dum_pop, ifstream& fin); 
  void init_pop(Pop &);
  virtual void function();
  virtual void f_output();
  void f_output(Pop&);
  void non_dominated_space(Pop &dum_pop);

  //the following functions are in connection with next_1_gene()
  virtual void next_1_gene();
  virtual void cross_over();
  virtual void mutation();
  virtual void selection();
  virtual void terminal_check();
  //The mentioned above functions are in connection with next_1_gene()
  void finish_write();
  virtual void all_open_file_close();
  inline void set_s_time()	{s_time=(get_time());}
  inline void set_e_time()	{e_time=(get_time());}
  inline void set_all_s_time()	{all_s_time=(get_time());}
  inline void set_all_e_time()        {all_e_time=(get_time());}
  inline int get_generation()	{return generation;}
  inline double get_e_s_time()	{return (e_time-s_time);}
  inline double get_all_e_s_time()	{return (all_e_time-all_s_time);}
  inline int get_select_pop_num() {return select.get_select_pop_num();}
  inline int get_first_pop_num() {return variable.pop_num;}
  //The mentioned above functions are in connection with cycle()
 public:
  One();
  virtual ~One() {};
  virtual void initGA();
  virtual void cycle();
};

#endif //_ONE_H_






