//-------------------------------------------------------------------
// FILENAME    : select.cpp
// DESCRIPTION : Selection 
//-------------------------------------------------------------------
#include <stdio.h>
#include <iostream>
#include <math.h>
#include <vector.h>
#include "select.h"
#include "gene.h"
#include "pop.h"
#include "oneisland.h"
#include "mtrand.h"
int Select::select_method=1;
int Select::obj_num=2;
int Select::select_pop_num=100;
int Select::sharing_need_pop=100;

//======================================
//class Select (select means selection of GA operators)
//======================================

void Select::selection(Pop& a)
{
  pop_num=a.get_pop_num();
  obj_setting(a);
  selection(a.ap_gene,pop_num);
  a.set_pop_num(pop_num);
}


/****************
selection has 3 type method.
1: roulet selection based on ranking only.
2: roulet selection based on ranking and sharing.
3: bi-tournamet selection based on ranking and sharing.
****************/

void Select::selection(Gene **ap_gene,int &num)
{
  pop_num=num;
  ranking_con(ap_gene,pop_num);
  if(select_method==1)
    roulet_select(ap_gene,pop_num);
  else if(select_method==2){
    ranking_square(ap_gene,pop_num);
    obj_sharing(ap_gene,pop_num,sharing_need_pop);
    roulet_select(ap_gene,pop_num);
  }
  else{
    ranking_square(ap_gene,pop_num);
    obj_sharing(ap_gene,pop_num,sharing_need_pop);
    bi_select_alpha(ap_gene,pop_num);
  }
  num=pop_num;
}

// ranking_con() define fitness value of all individuals base on ranking method only.
void Select::ranking_con(Pop& a)
{
  pop_num = a.get_pop_num();
  ranking_con(a.ap_gene,pop_num);
}

void Select::ranking_con(Gene **ap_gene,const int &dum)
{
  int i,j,s;
  double m,k,*rank,ptr;
  pop_num=dum;
  rank = new double[pop_num];
  
  ptr=0;

  for(i=0;i<pop_num;i++)
    rank[i]=1.0;
  
  for(i=0;i<pop_num;i++){
    for(s=i+1;s<pop_num;s++){
      m=0.0;
      k=0.0;
      for(j=0;j<obj_num;j++){
	if(ap_gene[i]->get_e(j)<ap_gene[s]->get_e(j))
	  m +=1.0;
	else if(ap_gene[i]->get_e(j)>ap_gene[s]->get_e(j))
	  k +=1.0;
	else if(ap_gene[i]->get_e(j)==ap_gene[s]->get_e(j)){
	  m += 1.0;
	  k += 1.0;
	}
      }
      if(m==(double)(obj_num)){
	rank[s] += 1;
      }
      else if(k==(double)(obj_num)){
	rank[i] += 1;
      }
    }
  }
  for(i=0;i<pop_num;i++){
    ap_gene[i]->get_f()=1.0/rank[i];
    //    printf("%d: fit %f\n",i,ap_gene[i]->get_f());
  }
  delete[] rank;
}

//roulet_select() is rulet selection method against all indiividual.
void Select::roulet_select(Gene **ap_gene,const int &num)
{
  int i,j,k,min_n[obj_num],dum;
  unsigned char check;
  double value,sum,d_sum,min_fx,dumy;
  vector<Gene> candicate_gene;
  candicate_gene.reserve(num);
  pop_num = num;

  for(i=0;i<pop_num;i++)
    ap_gene[i]->safe_flag=OUT;

  for(j=0;j<obj_num;j++){
    min_n[j]=0;
    min_fx=ap_gene[0]->get_c(j);
    for(i=1;i<pop_num;i++){
      if(min_fx>ap_gene[i]->get_c(j)){
	min_fx=ap_gene[i]->get_c(j);
	min_n[j]=i;
      }
    }
    ap_gene[min_n[j]]->safe_flag=SAFE;
  }
  sum=0.0;  
  dum=0;
  for(i=0;i<pop_num;i++){
    if(ap_gene[i]->safe_flag){
      candicate_gene.push_back(*ap_gene[i]);
      dum++;
    }
    sum += ap_gene[i]->get_f();
  }
  dum = pop_num-dum;

  for(i=0;i<dum;i++){
    dumy=genrand();
    value = dumy*sum;
    d_sum=0.0;
    check=0;
    for(k=0;k<pop_num;k++){
      d_sum += ap_gene[k]->get_f();
      if(value<=d_sum){
	//	cout<<k<<",";
	candicate_gene.push_back(*ap_gene[k]);
	check=1;
      }
      if(check)
	break;
    }
  }
  for(i=0;i<pop_num;i++){
    *ap_gene[i]=candicate_gene[i];
  }
}

// ranking_square() square the value of individuals fitness.
void Select::ranking_square(Gene **ap_gene,int &num)
{
  int i;
  pop_num=num;
  for(i=0;i<pop_num;i++)
    ap_gene[i]->get_f() = pow(ap_gene[i]->get_f(),2.0);
}

// obj_sharing define fitness value by using sharing function.
// obj_sharing is executed in objective area.

void Select::obj_sharing(Pop& a)
{
  int pop_num;
  pop_num = a.get_pop_num();
  obj_sharing(a.ap_gene,pop_num,sharing_need_pop);
}


void Select::obj_sharing(Gene **ap_gene,int &dum_pop_num,const int &dum)
{
  int i,j,k;
  double rou_shar,r[obj_num],max_fx[obj_num],min_fx[obj_num],distance,dis_2[obj_num],space,*S_h;
  
  pop_num=dum_pop_num;
  S_h = new double[pop_num];
  for(k=0;k<obj_num;k++){
    max_fx[k]=1;
    min_fx[k]=0;
    r[k]=(max_fx[k]-min_fx[k]);
  }
  space=0.0;
  for(k=0;k<obj_num;k++){
    space += pow(r[k],2.0);
  }
  space=sqrt(space);
  rou_shar=space/(double)dum;
  for(i=0;i<pop_num;i++){
    S_h[i]=1.0;
  }

  for(i=0;i<pop_num;i++){
    for(k=i+1;k<pop_num;k++){
      distance=0.0;
      for(j=0;j<obj_num;j++){
	dis_2[j] = pow((ap_gene[i]->get_c(j))-(ap_gene[k]->get_c(j)),2.0);
	distance += dis_2[j];
      }
      distance = sqrt(distance);
      if(distance<rou_shar){
	S_h[i] += 1.0-distance/rou_shar;
	S_h[k] += 1.0-distance/rou_shar;
      }
    }
  }
  for(i=0;i<pop_num;i++){
    ap_gene[i]->get_f()=ap_gene[i]->get_f()/S_h[i];
  }
  delete[] S_h;
}

// bi_select is tournamet selection method using two individuals.

void Select::bi_select(Pop &pop)
{
  pop_num=pop.get_pop_num();
  select_pop_num=pop_num;
  
  int i;
  for(i=0;i<pop_num;i++)
    pop.ap_gene[i]->safe_flag=OUT;
  bi_select_alpha(pop.ap_gene,pop_num);
  pop.set_pop_num(pop_num);
}

void Select::bi_select_alpha(Gene **ap_gene,int& dum_pop_num)
{
  bi_select_alpha(ap_gene,dum_pop_num,select_pop_num);
}

void Select::bi_select_alpha(Gene **ap_gene,int& dum_pop_num,const int& dum_dum)
{
  int j,hit[2];
  vector<Gene> candicate_gene;

  candicate_gene.reserve(dum_dum);
  pop_num=dum_pop_num;
  select_pop_num=dum_dum;
  for(j=0;j<select_pop_num;j++){
    hit[0]=normal_genrand_int(pop_num);
    hit[1]=normal_genrand_int(pop_num);

    if(ap_gene[hit[0]]->get_f() > ap_gene[hit[1]]->get_f()){
      candicate_gene.push_back(*ap_gene[hit[0]]);
    }
    else{
      candicate_gene.push_back(*ap_gene[hit[1]]);
    }
  }
  for(j=0;j<select_pop_num;j++)
    *ap_gene[j]=candicate_gene[j];
  dum_pop_num = select_pop_num;
}


// tournamet_select() is not pure tournamet selection method.
// This method selects individuals which have higher fitness value.

void Select::tournament_select(Pop& a)
{
  pop_num = a.get_pop_num();
  tournament_select(a.ap_gene,pop_num,select_pop_num);
  a.set_pop_num(pop_num);
}

void Select::tournament_select(Gene **ap_gene,int& dum_pop_num,const int& dum_dum)
{
  int i;
  pop_num = dum_pop_num;
  for(i=0;i<pop_num;i++)
    ap_gene[i]->safe_flag=OUT;
  tournament_select_alpha(ap_gene,dum_pop_num,dum_dum);
}

void Select::tournament_select_alpha(Gene **ap_gene,int& dum_pop_num)
{
  tournament_select_alpha(ap_gene,dum_pop_num,select_pop_num);
}

// tournament_select_alpha() is core function of tournament_select().

void Select::tournament_select_alpha(Gene **ap_gene,int& dum_pop_num,const int& dum_dum)
{
  int i;
  pop_num=dum_pop_num;
  f_qsort(ap_gene,0,pop_num-1);
  for(i=0;i<dum_dum;i++){
    ap_gene[i]->safe_flag=SAFE;
  }
  select_pop(ap_gene,pop_num);
  dum_pop_num=pop_num;
}

/***********************
select_pop() is the method for re arranging population.
if  individual that has safe_flag=OUT  remove from population.
 *********************/

void Select::select_pop(Pop& a)
{
  pop_num = a.get_pop_num();
  select_pop(a.ap_gene,pop_num);
  a.set_pop_num(pop_num);
}

void Select::select_pop(Gene **ap_gene,int &dum)
{
  int i,k;
  Gene ap_hit;
  
  pop_num=dum;
  for(i=0,k=0;i<pop_num;i++){
    if(ap_gene[i]->safe_flag){
      ap_hit = *ap_gene[i];
      *ap_gene[k]=ap_hit;
      k++;
    }
  }
  pop_num=k;
  dum = pop_num;
}


/*******************
obj_setting() sets current_obj value of population.
********************/

void Select::obj_setting(Pop &pop)
{
  int i,k;
  double max_fx[obj_num],min_fx[obj_num],value,r[obj_num];
  double dif;
  
  for(k=0;k<obj_num;k++){
    max_fx[k]=pop.ap_gene[0]->get_e(k);
    min_fx[k]=pop.ap_gene[0]->get_e(k);
    for(i=1;i<pop.get_pop_num();i++){
      value = pop.ap_gene[i]->get_e(k);
      if(max_fx[k]<value)
	max_fx[k]=value;
      else if(min_fx[k]>value)
	min_fx[k]=value;
    }
    r[k]=(max_fx[k]-min_fx[k]);
  }
  
  for(i=0;i<pop.get_pop_num();i++)
    for(k=0;k<obj_num;k++){
	  dif = pop.ap_gene[i]->get_e(k)-min_fx[k];
	  if(dif!=0 || r[k]!=0)
		pop.ap_gene[i]->get_c(k)=(pop.ap_gene[i]->get_e(k)-min_fx[k])/r[k];
	}

  for(k=0;k<obj_num;k++){
    pop.set_max_obj(k)=max_fx[k];
    pop.set_min_obj(k)=min_fx[k];
  }
}


void Select::sort_pop(Pop& a)
{
  new_sort(a,0);
}	


void Select::new_sort(Pop& a,int base_obj)
{
  pop_num=a.get_pop_num();
  qsort(a.ap_gene,0,pop_num-1,base_obj);
}


/*******************
f_sort() sorts population based on fitness value of each individuals.
 ******************/
void Select::f_sort(Pop& a)
{
  pop_num=a.get_pop_num();
  f_qsort(a.ap_gene,0,pop_num-1);
}

void Select::reduce_same_pop(Pop& a)
{
  pop_num = a.get_pop_num();
  reduce_same_pop(a.ap_gene,pop_num);
  a.set_pop_num(pop_num);
}

void Select::reduce_same_pop(Gene **ap_gene,int& dum_num)
{
  int i,k;
  dum_num=pop_num;
  for(i=0;i<pop_num;i++)
    ap_gene[i]->safe_flag=SAFE;
  
  for(i=0;i<pop_num;i++){
    if(!ap_gene[i]->safe_flag)
      continue;
    for(k=i+1;k<pop_num;k++){
      if(!ap_gene[k]->safe_flag)
	continue;
      if(*ap_gene[i] == *ap_gene[k])
	ap_gene[k]->safe_flag=OUT;
    }
  }
  select_pop(ap_gene,dum_num);
}




void Select::only_pareto_check(Gene **ap_gene,const int& dum_num)
{
  int i,s,j,m,k;
  pop_num=dum_num;
  for(i=0;i<pop_num;i++){
    ap_gene[i]->safe_flag=SAFE;
  }
  for(i=0;i<pop_num;i++){
    if(!ap_gene[i]->safe_flag)
      continue;
    for(s=i+1;s<pop_num;s++){
      if(!ap_gene[s]->safe_flag)
	continue;
      k=0;
      m=0;
      for(j=0;j<obj_num;j++){
	if(ap_gene[i]->get_e(j) < ap_gene[s]->get_e(j))
	  m += 1;
	else if(ap_gene[i]->get_e(j) > ap_gene[s]->get_e(j))
	  k += 1;
	else if(ap_gene[i]->get_e(j) == ap_gene[s]->get_e(j)){
	  m += 1;
	  k += 1;
	}
      }
      
      if(m==(obj_num))
	ap_gene[s]->safe_flag=OUT;
      else if(k==(obj_num))
	ap_gene[i]->safe_flag=OUT;
      if(!ap_gene[i]->safe_flag)
	break;
    }
  }
}

void Select::only_pareto_select(Pop &pop)
{
  pop_num = pop.get_pop_num();
  only_pareto_check(pop.ap_gene,pop_num);
  select_pop(pop.ap_gene,pop_num);
  pop.set_pop_num(pop_num);
}


void f_inssort(Gene **ap_gene,int first,int end)
{
  int i, j;
  Gene temp;
  for(i = 1 + first; i < end; i++)
    {
      temp = *ap_gene[i];
      for(j = i - 1;
          j >= first && ap_gene[j]->get_f()< temp.get_f(); j--)
                                *ap_gene[j + 1] = *ap_gene[j];
                        *ap_gene[j + 1] = temp;
                }
}

void f_qsort(Gene **apGene,int first,int end)
{
  if(end - first <= 10)
    f_inssort(apGene, first, end);
  int i, j;
  double val;
  Gene temp;

  val = apGene[(first + end) / 2]->get_f();

  i = first; j = end;
  for(;;)
    {
      while(apGene[i]->get_f() > val) i++;
      while(val > apGene[j]->get_f()) j--;
      if(i >= j) break;
      temp = *apGene[i]; *apGene[i] = *apGene[j]; *apGene[j] = temp;
      i++; j--;
    }
  if(first < i - 1) f_qsort(apGene, first, i - 1);
  if(j + 1 < end)   f_qsort(apGene, j + 1, end);
}



void inssort(Gene **ap_gene,int first,int end,const int& base_obj)
{
  int i, j;
  Gene temp;
  
  //	printf("start inssort\n");
		for(i = 1 + first; i < end; i++)
		  {
			temp = *ap_gene[i];
			for(j = i - 1;
				j >= first && ap_gene[j]->get_e(base_obj)< temp.get_e(base_obj); j--)
				*ap_gene[j + 1] = *ap_gene[j];
			*ap_gene[j + 1] = temp;
		}
	}

	// quick sort (decremental direction)
	void qsort(Gene **apGene,int first,int end,const int& base_obj)
	{
		// quick sort is slower than insertion sort when range is too short.
	//	printf("start qsort base_obj %d\n",base_obj);
		if(end - first <= 10)
			inssort(apGene, first, end,base_obj);

		int i, j;
		double val;
		Gene temp;

		val = apGene[(first + end) / 2]->get_e(base_obj);

		i = first; j = end;
		for(;;)
		{
			while(apGene[i]->get_e(base_obj) > val) i++;
			while(val > apGene[j]->get_e(base_obj)) j--;
			if(i >= j) break;

			temp = *apGene[i]; *apGene[i] = *apGene[j]; *apGene[j] = temp;
			i++; j--;
		}
		if(first < i - 1) qsort(apGene, first, i - 1,base_obj);
		if(j + 1 < end)   qsort(apGene, j + 1, end,base_obj);
	}

	void marge(Gene **ap_gene,int first,int end,const int& base_obj)
	{
		int i,j,k,p,middle;
		Gene *work;

		middle = (first+end)/2;

		work = new Gene[middle+2];
		p=0;
		for(i=first;i<=middle;i++)
			work[p++]=*ap_gene[i];
		i=middle+1; j=0; k=first;

		while(i<end&&j<p){
			if(work[j].get_e(base_obj) >= ap_gene[i]->get_e(base_obj))
				*ap_gene[k++]=work[j++];
			else
				*ap_gene[k++]=*ap_gene[i++];

		}

		while(j<p)
			*ap_gene[k++]=work[j++];
		delete [] work;
	}

	void marge_h(Gene **ap_gene,int first,int end,const int& base_obj,const u8 &flag)
{
		int i,j,k,p,middle,dum;
		Gene *work;
		middle = (first+end)/2;

		work = new Gene[middle+2];
	p=0;
	for(i=first;i<=middle;i++)
		work[p++]=*ap_gene[i];
	i=middle+1; j=0; k=first;dum=0;

	if(flag==0){  // "0"  is lower falf 
		while(i<=end && j<p){
			if(work[j].get_e(base_obj) >= ap_gene[i]->get_e(base_obj))
				j++;
			else
				i++;
			dum++;
			if(dum > middle)
				break;
		}

		while(i<=end&&j<p){
			if(work[j].get_e(base_obj) >= ap_gene[i]->get_e(base_obj))
				*ap_gene[k++]=work[j++];
			else
				*ap_gene[k++]=*ap_gene[i++];
		}
		while(j<p)
			*ap_gene[k++]=work[j++];
		while(i<=end)
			*ap_gene[k++]=*ap_gene[i++];
	}
	else{ // "1" is upper half.

		while(i<=end&&j<=p){
			if(work[j].get_e(base_obj) >= ap_gene[i]->get_e(base_obj))
				*ap_gene[k++]=work[j++];
			else
				*ap_gene[k++]=*ap_gene[i++];
			dum ++;
			if(dum>middle)
				break;
		}
	}
	delete [] work;
}

void inssort_int(Gene **ap_gene,int first,int end,int *base)
{
  int i,j,b_tmp;
  Gene temp;
  
  for(i = 1 + first; i < end; i++)
    {
      temp = *ap_gene[i];
      b_tmp=base[i];
      for(j = i - 1;
	  j >= first && base[j]< b_tmp; j--){
	*ap_gene[j + 1] = *ap_gene[j];
	base[j+1]=base[j];
      }
      *ap_gene[j + 1] = temp;
      base[j+1] = b_tmp;
    }

}

// quick sort (decremental direction)
void qsort_int(Gene **apGene,int first,int end,int *base)
{
  if(end - first <= 10)
    inssort_int(apGene, first, end,base);
  
  int i, j,b_tmp;
  double val;
  Gene temp;
  
  val = base[(first + end) / 2];
  
  i = first; j = end;
  for(;;)
    {
      while(base[i] > val) i++;
      while(val > base[j]) j--;
      if(i >= j) break;

      temp = *apGene[i];
      b_tmp=base[i];
      *apGene[i] = *apGene[j];
      base[i]=base[j];
      *apGene[j] = temp;
      base[j]=b_tmp;
      i++; j--;
    }
  if(first < i - 1) qsort_int(apGene, first, i - 1,base);
  if(j + 1 < end)   qsort_int(apGene, j + 1, end,base);
}

void sort(int *apInt,int first,int end)
{
  qsort(apInt,first,end);
}

void qsort(int *apInt,int first,int end)
{
  // quick sort is slower than insertion sort when range is too short.
  if(end - first <= 10)
    inssort(apInt, first, end);
  
  int i, j;
  double val;
  int temp;
  
  val = apInt[(first + end) / 2];
  
  i = first; j = end;
  for(;;)
    {
      while(apInt[i] > val) i++;
      while(val > apInt[j]) j--;
      if(i >= j) break;
      
      temp = apInt[i]; apInt[i] = apInt[j]; apInt[j] = temp;
      i++; j--;
    }
  if(first < i - 1) qsort(apInt, first, i - 1);
  if(j + 1 < end)   qsort(apInt, j + 1, end);
}

void inssort(int *ap_gene,int first,int end)
{
  int i, j;
  int temp;
  
  for(i = 1 + first; i < end; i++)
    {
      temp = ap_gene[i];
      for(j = i - 1;
	  j >= first && ap_gene[j]< temp; j--)
	ap_gene[j + 1] = ap_gene[j];
      ap_gene[j + 1] = temp;
    }
}


void f_inssort(vector<Gene>& ap_gene,int first,int end)
{
  int i, j;
  Gene temp;
  for(i = 1 + first; i < end; i++)
    {
      temp = ap_gene[i];
      for(j = i - 1;
          j >= first && ap_gene[j].get_f()< temp.get_f(); j--)
                        ap_gene[j + 1] = ap_gene[j];
                        ap_gene[j + 1] = temp;
                }
}

void f_qsort(vector<Gene>& apGene,int first,int end)
{
  if(end - first <= 10)
    f_inssort(apGene, first, end);
  int i, j;
  double val;
  Gene temp;

  val = apGene[(first + end) / 2].get_f();

  i = first; j = end;
  for(;;)
    {
      while(apGene[i].get_f() > val) i++;
      while(val > apGene[j].get_f()) j--;
      if(i >= j) break;
      temp = apGene[i]; apGene[i] = apGene[j]; apGene[j] = temp;
      i++; j--;
    }
  if(first < i - 1) f_qsort(apGene, first, i - 1);
  if(j + 1 < end)   f_qsort(apGene, j + 1, end);
}



void inssort(vector<Gene>& ap_gene,int first,int end,const int& base_obj)
{
  int i, j;
  Gene temp;
  
  //	printf("start inssort\n");
  for(i = 1 + first; i < end; i++)
    {
      temp = ap_gene[i];
      for(j = i - 1;
	  j >= first && ap_gene[j].get_e(base_obj)< temp.get_e(base_obj); j--)
	ap_gene[j + 1] = ap_gene[j];
      ap_gene[j + 1] = temp;
    }
}

// quick sort (decremental direction)
void qsort(vector<Gene>& apGene,int first,int end,const int& base_obj)
{
  // quick sort is slower than insertion sort when range is too short.
  //	printf("start qsort base_obj %d\n",base_obj);
  if(end - first <= 10)
    inssort(apGene, first, end,base_obj);
  
  int i, j;
  double val;
  Gene temp;
  
  val = apGene[(first + end) / 2].get_e(base_obj);
  
  i = first; j = end;
  for(;;)
    {
      while(apGene[i].get_e(base_obj) > val) i++;
      while(val > apGene[j].get_e(base_obj)) j--;
      if(i >= j) break;
      
      temp = apGene[i]; apGene[i] = apGene[j]; apGene[j] = temp;
      i++; j--;
    }
  if(first < i - 1) qsort(apGene, first, i - 1,base_obj);
  if(j + 1 < end)   qsort(apGene, j + 1, end,base_obj);
}















