#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <netinet/in.h>
#include <vector.h>
#include "gene.h"
#include "base.h"

using namespace std;

int Gene::Rna_size=0;
vector<double> Gene::variable_min;
vector<double> Gene::variable_max;
int Gene::real_chrom=0;
int Gene::plan_num=0;
int Gene::obj_num=0;

Gene::Gene()
{
	safe_flag = SAFE;
	change_flag = CHANGE;
	
	if(real_chrom)
	  locus.resize(real_chrom);
	if(obj_num){
	  obj.eval_value = new double[obj_num];
 	 obj.const_eval = new double[obj_num];
 	}
}

Gene::Gene(const int& x)
{
  if(!real_chrom){
    real_chrom=x;
    plan_num=x;
  }
  locus.resize(real_chrom);
  if(!variable_min.size()){
    variable_min.resize(plan_num);
    variable_max.resize(plan_num);
  }
  safe_flag = SAFE;
  change_flag = CHANGE;
}

Gene::Gene(const int& x,const int& y)
{
  if(!obj_num){
    obj_num=y;
    obj.set_obj_num(obj_num);
  }
  if(!real_chrom)
    real_chrom = x;

  locus.resize(real_chrom);
  obj.eval_value = new double[obj_num];
  obj.const_eval = new double[obj_num];
  safe_flag = SAFE;
  change_flag = CHANGE;
}

Gene::Gene(const Gene& f)
{
  locus.resize(real_chrom);
  if(obj.eval_value==NULL){
    obj.eval_value = new double[obj_num];
    obj.const_eval = new double[obj_num];
  }
  copy(f);
}

Gene::~Gene()
{

}

double Gene::get_locus_value(const int& point) const
{
  unsigned int i_start,i_end,size;
  double dum;
  size=(real_chrom)/plan_num;
  i_start = (size)*point;
  i_end = (size)*(point+1);
  dum= (double)get_variable_value(i_start,i_end);
  dum /= (1LL << (size));
  
  dum = dum*(variable_max[point]-variable_min[point]) + variable_min[point];
  return dum;
}


double Gene::get_locus_value_p(const int& point) const
{
  unsigned int i_start,i_end,size;
  double dum;
  size=(real_chrom)/plan_num;
  i_start = (size)*point;
  i_end = (size)*(point+1);
  dum= (double)get_variable_value(i_start,i_end);

  if(point ==0)
  	cout<<"(in get)dum "<<dum<<endl;

  dum /= (1LL << (size));
  
  dum = dum*(variable_max[point]-variable_min[point]) + variable_min[point];
  return dum;
}

double Gene::get_variable_value(const int& start,const int& end) const
{
	unsigned int i;
	long long real, mask;
	bool sw = false;
	for(i = start,real = 0; i < end; i++)
	  {
	    if(locus[i].locus_r){
		  mask = 1LL << (end - i - 1);	
	      real |= mask;
	      
	    }
	  }
	
	for(mask = 1LL << ((end-start) - 1); mask != 0; mask >>= 1LL)
	  {
	  if(sw)
	    {
	      if(real & mask) sw = false;
	      real ^= mask;
	    }
	  else if(real & mask) sw = true;
	  }
	return (double)real;
}

void Gene::set_locus_value(const double *value) 
{
  unsigned int i_start,i_end,size;
  double dum;
  long long num;
  size=(real_chrom)/plan_num;
  
  for(int i=0;i<plan_num;i++){
    dum  = (value[i]-variable_min[i])/(variable_max[i]-variable_min[i]);
    dum *=  (1LL << (size));
    num = static_cast<long long> (dum + 0.5);
    i_start = (size)*i;
    i_end = (size)*(i+1);
    set_variable_value(i_start,i_end,num);
  }
}

void Gene::set_locus_value(const vector <double> value) 
{
  unsigned int i_start,i_end,size;
  double dum;
  long long num;
  size=(real_chrom)/plan_num;
  for(int i=0;i<plan_num;i++){
    dum  = (value[i]-variable_min[i])/(variable_max[i]-variable_min[i]);
    dum *=  (1LL << (size));
    num = static_cast<long long> (dum + 0.5);
    i_start = (size)*i;
    i_end = (size)*(i+1);
    set_variable_value(i_start,i_end,num);
  }
}


void Gene::set_locus_value(const int &point,const double& value) 
{
  unsigned int i_start,i_end,size;
  double dum;
  long long num;
  size=(real_chrom)/plan_num;
  
  dum  = (value-variable_min[point])/(variable_max[point]-variable_min[point]);

// dum *=  (1LL << (size-1));
  dum *=  (1LL << (size));
  num = static_cast<long long> (dum + 0.5);
  
  cout<<"raw num = "<< num<<endl;
  
  i_start = (size)*point;
  i_end = (size)*(point+1);
  set_variable_value(i_start,i_end,num);
}

void Gene::set_variable_value(const int& start,const int& end,long long num)
{
  long long mask;
  bool f = false;

  for(mask = 1LL << (end-start); mask != 0; mask >>= 1LL)
  {
    if(f)
      {
	num ^= mask;
	if(num & mask) f = false;
      }
    else
      if(num & mask) f = true;
  }
  
  int i = end;
  while(start < i){
    locus[--i].locus_r = num & 1;
    num >>= 1LL;
  }
}


void Gene::set_parameter(const int& x,const int& y,const int& z)
{
  if(!obj_num)
    set_obj_num(x);
  if(!plan_num)
    plan_num=y;
  if(!real_chrom)
    real_chrom=z;
  if(!variable_min.size()){
    variable_min.resize(plan_num);
    variable_max.resize(plan_num);
  }
}

void Gene::set_func_num(const int& func_num)
{
  int i;
  if(func_num==1){
    variable_min[0]=1.0;
    variable_min[1]=1.0;
    variable_max[0]=4.0;
    variable_max[1]=2.0;
  }
  else if(func_num==2){
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=7.0;
    }
  }
  else if(func_num==3){
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=1.0;
    }
  }
  else if(func_num==4){
      variable_min[0]=-3.5;
      variable_max[0]=0.0;
      variable_min[1]=-3.5;
      variable_max[1]=1.5;
    }
  else if(func_num==5){
      variable_min[0]=1.0;
      variable_max[0]=4.0;
      variable_min[1]=1.0;
      variable_max[1]=2.0;
  }
  else if(func_num==6){
      variable_min[0]=0.0;
      variable_max[0]=1.0;
      for(i=1;i<plan_num;i++){
	variable_min[i]=-30.0;
	variable_max[i]=30.0;
      }
  }
  else if(func_num==7){
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=1.0;
    }
  }

  else if(func_num==8){
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=1.0;
    }
  }
  else if(func_num==9){
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=1.0;
    }
  }
  else if(func_num==10){
    variable_min[0]=0.0;
    variable_max[0]=1.0;
    for(i=1;i<plan_num;i++){
      variable_min[i]=-5.0;
      variable_max[i]=5.0;
    }
  }
  else if(func_num==11){
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=1.0;
    }
  }
  else if(func_num==12){
    for(i=0;i<plan_num;i++){
      variable_min[i]=-1000.0;
      variable_max[i]=1000.0;
    }
  }
  else if(func_num==13){
    for(i=0;i<plan_num;i++){
      variable_min[i]=-5.0;
      variable_max[i]=5.0;
    }
  }
  else{
    for(i=0;i<plan_num;i++){
      variable_min[i]=0.0;
      variable_max[i]=7.0;
    }
  }
}

Gene Gene::init(const int& x,const int& y)
{
  if(!obj_num){
    obj_num=y;
    obj.set_obj_num(obj_num);
  }
  if(!real_chrom)
    real_chrom = x;
  locus.resize(real_chrom);  
  obj.eval_value = new double[obj_num];
  obj.const_eval = new double[obj_num];
  safe_flag = SAFE;
  change_flag = CHANGE;
  return *this;
}

void Gene::copy(const Gene& f)
{
  int i;
  for(i=0;i<real_chrom;i++)
    locus[i]=f.locus[i];
  obj.copy(f.obj);
  safe_flag=f.safe_flag;
  change_flag = f.change_flag;
}

Gene Gene::operator=(const Gene& f)
{
  locus.clear();
  locus.resize(real_chrom);
  if(obj.eval_value==NULL){
    obj.eval_value = new double[obj_num];
    obj.const_eval = new double[obj_num];
  }
  copy(f);
  return *this;
}

bool Gene::operator==(const Gene &right)
{
  int i,k;
  k=0;
  for(i=0;i<real_chrom;i++){
    if(locus[i]==right.locus[i])
      k++;
  }
  if(k==real_chrom)
    return true;
  else
    return false;
}

inline void Gene::printf_para()
{
	printf("obj_num =%d  plan_num=%d  real_chrom = %d \n",obj_num,plan_num,real_chrom);
}


void Gene::show_chromosome()
{
	int i;
	obj.Obj_printf();
	for(i=0;i<plan_num;i++){
		if((i+1)%5||i==0)
		  printf("%f,",get_locus_value(i));
		else{
		  cout<<"\n";
		  printf("%f,",get_locus_value(i));
		}
	}
	printf("\n");
}

void Gene::f_show_obj(FILE *fp)
{
	int i;
//	fprintf(fp,"[");
	for(i=0;i<obj_num;i++){
		if(i!=(obj_num-1))
			 fprintf(fp,"%f\t",obj[i]);
		else
			fprintf(fp,"%f\n",obj[i]);
	}
}


void Gene::f_chromosome(FILE *fp)
{
	int i;
        for(i=0;i<real_chrom;i++){
	  if(((i+1)%5)||i==0)
	    locus[i].f_out(fp);
	  else{
	    fprintf(fp,"\n");
	    locus[i].f_out(fp);
	  }
        }
        fprintf(fp,"\n");
}



void Gene::f_show_plan_value(FILE *fp)
{
	int i;
	f_show_obj(fp);
	for(i=0;i<plan_num;i++){
		if((i+1)%5||i==0)
		  fprintf(fp,"%f,",get_locus_value(i));
		else{
		  fprintf(fp,"\n%f,",get_locus_value(i));
		}
	}
	fprintf(fp,"\n");
}

int Gene::GetRnaLength()
{
  if(Rna_size)
    return Rna_size;
  else{
    Rna_size=GetRnaLength_set();
    return Rna_size;
  }    
}

int Gene::GetRnaLength_set()
{
        Rna_size=0;
        Rna_size += (sizeof(double))*obj_num;
        Rna_size += (sizeof(double))*obj_num;
	Rna_size += real_chrom*(sizeof(u8));
        return (Rna_size);
}

const u8* Gene::transcription(u8 *buffer) 
{
  int i,a,n;
  i32 mant;
  i32 mant_n,exp2;

  a=0;
  for(i=0;i<obj_num;i++){
    mant = (i32)(frexp(obj[i],&n) * (1 << 30));
    mant_n = htonl(mant);
    exp2   = htonl(n);
    memcpy(buffer + a, &mant_n, 4);
    memcpy(buffer + a + 4,&exp2, 4);
    a += 8;
  }
  a= obj_num*8;

  for(i=0;i<obj_num;i++){
    mant = (i32)(frexp(obj.get_c_eval(i),&n) * (1 << 30));
    mant_n = htonl(mant);
    exp2   = htonl(n);
    memcpy(buffer + a, &mant_n, 4);
    memcpy(buffer + a + 4,&exp2, 4);
    a += 8;
  }
  for(i=0;i<real_chrom;i++){
    buffer[a]=locus[i].locus_r;
    a += 1;
  }
  return buffer;
}

void Gene::reverse_transcription(const u8 *data)
{
	int n,i,a;
	double dum;
	i32 mant;

	safe_flag=SAFE;
	change_flag=NOCHANGE;
	a=0;

	for(i=0;i<obj_num;i++){
	  mant = ntohl(*(i32 *)(data+a));
	  n = ntohl(*(i32 *)(data + a + 4));
	  dum = ldexp((double)mant, n - 30);
	  obj[i] = dum;
	  a += 8;
	}
	a = 8*obj_num;

	for(i=0;i<obj_num;i++){
	  mant = ntohl(*(i32 *)(data+a));
	  n = ntohl(*(i32 *)(data + a + 4));
	  dum = ldexp((double)mant, n - 30);
	  obj.get_c_eval(i) = dum;
	  a += 8;
	}

	locus.reserve(real_chrom);
	locus.resize(real_chrom);
	for(i=0;i<real_chrom;i++){
	  locus[i].locus_r = data[a];
	  a +=1;
	}
}



void Gene::reverse_transcription2(const u8 *data)
{
	int n,i,a;
	double dum;
	i32 mant;

	safe_flag=SAFE;
	change_flag=CHANGE;
	a=0;

	for(i=0;i<obj_num;i++){
	  mant = ntohl(*(i32 *)(data+a));
	  n = ntohl(*(i32 *)(data + a + 4));
	  dum = ldexp((double)mant, n - 30);
	  obj[i] = dum;
	  //	  printf("obj[%d] %f\n",i,obj[i]);
	  //	  cout<<"obj["<<i<<"]="<<dum<<"\n";
	  a += 8;
	}
	a = 8*obj_num;

	for(i=0;i<obj_num;i++){
	  mant = ntohl(*(i32 *)(data+a));
	  n = ntohl(*(i32 *)(data + a + 4));
	  dum = ldexp((double)mant, n - 30);
	  obj.get_c_eval(i) = dum;
	  a += 8;
	}

	locus.reserve(real_chrom);
	locus.resize(real_chrom);

	for(i=0;i<real_chrom;i++){
	  locus[i].locus_r = data[a];
	  a +=1;
	}
}


Chromo Chromo::operator=(const Chromo& f)
{
  locus_r=f.locus_r;
  return *this;
}

Chromo Chromo::operator=(const int& f)
{
	locus_r=f;
	return *this;
}


bool Chromo::operator==(const Chromo& f)
{
  if(locus_r==f.locus_r)
    return true;
  else
    return false;
}

void Chromo::f_out(FILE *fp)
{
        fprintf(fp,"%d  ",locus_r);
}

u8 Chromo::operator^(const Chromo& dum)
{
  return (locus_r^dum.locus_r);
}

int Obj::obj_num=0;

Obj::Obj()
{
	eval_value=NULL;
	const_eval=NULL;
	fitness_value=0.0;
}

Obj::Obj(int x)
{
	obj_num=x;
   	int i;
  	eval_value = new double[obj_num];
	const_eval = new double[obj_num];
     	fitness_value=0.0;
  	for(i=0;i<obj_num;i++)
        eval_value[i]=0.0;
}

Obj::Obj(const Obj &f)
{
  if(eval_value==NULL){
    eval_value = new double[obj_num];	
    const_eval = new double[obj_num];
  }
	copy(f);
}

Obj::~Obj()
{
  if(eval_value != NULL){
    delete[] eval_value;
    delete[] const_eval;
  }
}
	
void Obj::copy(const Obj& f)
{
	size_t n;
	n = sizeof(double)*(obj_num);
	memcpy(eval_value,f.eval_value,n);
	memcpy(const_eval,f.const_eval,n);
	fitness_value=f.fitness_value;
}

void Obj::Obj_printf()
{
	int i;
	printf("eval_value:");

	//	cout<<"in Obj obj_num="<<obj_num<<"\n";

	for(i=0;i<obj_num;i++){
		if(i%5||i==0)
			printf("[%d]%f  ",i,eval_value[i]);	
		else
			printf("[%d]%f\n",i,eval_value[i]);	
	}
	printf("\n");
}


void Obj::Obj_printf_curent()
{
	int i;
	printf("const_eval:");
	for(i=0;i<obj_num;i++){
		if(i%5||i==0)
			printf("[%d]%f  ",i,const_eval[i]);	
		else
			printf("[%d]%f\n",i,const_eval[i]);	
	}
	printf("\n");
}

Obj Obj::operator=(const Obj& f)
{
  if(eval_value != NULL){
    delete [] eval_value;
    delete [] const_eval;
  }
  eval_value = new double[obj_num];
  const_eval = new double[obj_num];
  copy(f);
  return *this;
}

u8 Obj::operator<(const Obj &dum)
{
  if(fitness_value<dum.fitness_value)
    return 1;
  else
    return 0;
}

u8 Obj::operator>(const Obj &dum)
{
  if(fitness_value>dum.fitness_value)
    return 1;
  else
    return 0;
}

u8 Obj::operator<=(const Obj &dum)
{
  if(fitness_value<=dum.fitness_value)
    return 1;
  else
    return 0;
}

u8 Obj::operator>=(const Obj &dum)
{
  if(fitness_value>=dum.fitness_value)
    return 1;
  else
    return 0;
}

